from etl_bq_tools.utils.time_execution import get_time_function_execution


def bq_set_project(project_id=None, logging=None):
    """
    :param project_id: String
    :param logging: Object
    :return:
    """
    import os

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    try:
        os.system(f"gcloud config set project {project_id}")
        logging.info(f"project update {project_id} ")
    except:
        logging.info(f"Project Not Found: {project_id} ")


@get_time_function_execution
def bq_execute_sql(sql=None, parameter=None, logging=None):
    """
    :param sql: file.sql
    :param parameter: {"parameter:[{"variable": "", "type": "", "value":""}]}
    :param logging: object
    :return:
    """

    import os

    if not sql:
        raise Exception('require var sql: {file.sql} ')

    file_sql = os.path.basename(sql)

    try:
        _parameter = ""
        if parameter:
            for key, values in parameter.items():
                for value in values:
                    _var = value["variable"]
                    _type = value["type"]
                    _val = value["value"]
                    _parameter = f"--parameter={_var}:{_type}:{_val}"
                    _parameter += f" {_parameter} \\ "

            sentence_bq = f"bq query --quiet  --use_legacy_sql=false {_parameter} --flagfile={sql} "
        else:
            sentence_bq = f"bq query --quiet  --use_legacy_sql=false --flagfile={sql} "
        os.system(sentence_bq)
        logging.info(f"Successfully: {file_sql}")
    except Exception as e:
        logging.info(f"Error: {file_sql}")



from etl_bq_tools.utils.time_execution import get_time_function_execution


@get_time_function_execution
def bq_create_dataset(project_id=None,
                      dataset_id=None,
                      logging=None,
                      key_path=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param key_path: file.json
    :param logging: object
    :return:
    """
    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id:{dataset_id} ')

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    dataset_id = f"{project_id}.{dataset_id}"
    dataset = bigquery.Dataset(dataset_id)
    dataset.location = "US"

    try:
        dataset = client.create_dataset(dataset, exists_ok=True)
        logging.info(f"Created dataset {project_id}.{dataset.dataset_id}")
    except Conflict:
        logging.info(f"Already Exists or ProjectID Not Found {project_id}.{dataset.dataset_id}")


@get_time_function_execution
def bq_delete_dataset(project_id=None,
                      dataset_id=None,
                      logging=None,
                      key_path=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery
    from google.cloud.exceptions import Conflict

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id:{dataset_id} ')

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    dataset_id = f"{project_id}.{dataset_id}"

    try:
        client.delete_dataset(dataset_id, delete_contents=True, not_found_ok=True)
        logging.info("Deleted dataset '{}'.".format(dataset_id))
    except Conflict:
        logging.info(f"Dataset Not Found {project_id}.{dataset_id}")


@get_time_function_execution
def bq_list_dataset(project_id=None,
                    logging=None,
                    key_path=None):
    """
    :param project_id: String
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    datasets = list(client.list_datasets())

    data_list = list()
    if datasets:
        logging.info(f"Datasets in project {project_id}:")
        for dataset in datasets:
            data_dict = dict()
            data_dict["project_id"] = dataset.project
            data_dict["dataset_id"] = dataset.dataset_id
            data_list.append(data_dict)
            print(f"\t {dataset.project}.{dataset.dataset_id}")
    else:
        logging.info(f"{project_id} project does not contain any datasets.")

    return data_list


@get_time_function_execution
def bq_create_table(project_id=None,
                    dataset_id=None,
                    table_id=None,
                    schema=None,
                    logging=None,
                    key_path=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param table_id: String
    :param schema: dict
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id: {dataset_id} ')
    if not table_id:
        raise Exception('require var table_id: {table_id} ')
    if not schema:
        raise Exception('require var schema ')

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    table_id = f"{project_id}.{dataset_id}.{table_id}"
    table = bigquery.Table(table_id, schema=schema)

    try:
        table = client.create_table(table)
        logging.info(f"Created Table {table.project}.{table.dataset_id}.{table.table_id}")
    except Conflict:
        logging.info(f"Already Exists or TableID Not Found {table.project}.{table.dataset_id}.{table.table_id}")


@get_time_function_execution
def bq_delete_table(project_id=None,
                    dataset_id=None,
                    table_id=None,
                    logging=None,
                    key_path=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param table_id: String
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id:{dataset_id} ')
    if not table_id:
        raise Exception('require var table_id: {table_id} ')

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    table_id = f"{project_id}.{dataset_id}.{table_id}"
    try:
        client.delete_table(table_id, not_found_ok=True)
        logging.info("Deleted table '{}'.".format(table_id))
    except Conflict:
        logging.info(f"Table Not Found {project_id}.{dataset.dataset_id}")


@get_time_function_execution
def bq_list_tables(project_id=None,
                   dataset_id=None,
                   logging=None,
                   key_path=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id:{dataset_id} ')

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    dataset_id = f"{project_id}.{dataset_id}"
    tables = client.list_tables(dataset_id)

    data_list = list()

    if tables:
        logging.info("Tables contained in '{}':".format(dataset_id))
        for table in tables:
            data_dict = dict()
            data_dict["project_id"] = table.project
            data_dict["dataset_id"] = table.dataset_id
            data_dict["table_id"] = table.table_id
            data_list.append(data_dict)
            print(f"\t {table.project}.{table.dataset_id}.{table.table_id}")
    else:
        logging.info(f"{project_id}.{dataset_id} dataset does not contain any tables.")
    return data_list


@get_time_function_execution
def bq_copy_table(orig_table_id=None,
                  dest_table_id=None,
                  logging=None,
                  key_path=None):
    """
    :param orig_table_id: String {project_id}.{dataset_id}.{table_id}
    :param dest_table_id: String {project_id}.{dataset_id}.{table_id}
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict

    if not orig_table_id:
        raise Exception('require var orig_table_id: {project_id}.{dataset_id}.{table_id} ')
    if not dest_table_id:
        raise Exception('require var dest_table_id: {project_id}.{dataset_id}.{table_id} ')

    client = bigquery.Client()
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)
    job_config = bigquery.CopyJobConfig()
    try:
        job = client.copy_table(orig_table_id, dest_table_id, job_config=job_config)
        job.result()
        logging.info(f"Copy Table {orig_table_id} from {dest_table_id}  ")
    except Conflict:
        logging.info(f"Table Not Found")


@get_time_function_execution
def bq_query_to_df(
        project_id=None, sql=None, file=None, use_legacy_sql=False,
        logging=None, key_path=None):
    """
    :param project_id: String
    :param sql: String
    :param file: String file_path.sql
    :param use_legacy_sql: Boolean
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not sql and not file:
        raise Exception('Required var sql or file')
    if sql and file:
        raise Exception('Only one variable is required {sql} or {file}')
    if file:
        if not str(file).endswith(".sql"):
            raise Exception('the file only support extension {file_path}.sql')

    client = bigquery.Client(project=project_id)
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    sql_text = ""
    if sql and not file:
        sql_text = sql

    if file and not sql:
        sql_text = open(file).read()

    try:
        query_config = bigquery.QueryJobConfig(use_legacy_sql=use_legacy_sql)
        df = client.query(sql_text, job_config=query_config).to_dataframe()
        logging.info("Successfully Query")
        return df
    except Exception as e:
        logging.info("Error Query")



from etl_bq_tools.utils.time_execution import get_time_function_execution


@get_time_function_execution
def upload_df_to_bq(
        project_id, dataset_id, table_id, key_path,
        schema=None, df=None, if_exists="truncate", logging=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param table_id: String
    :param key_path: file.json
    :param schema: dict
    :param df: Dataframe
    :param if_exists: String -> "append" or "truncate"
    :param logging: object
    :return:
    """
    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict
    import pandas as pd

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id: {dataset_id} ')
    if not table_id:
        raise Exception('require var table_id: {table_id} ')
    if not isinstance(df, pd.DataFrame):
        raise Exception('require var df: {df} ')
    if not if_exists:
        raise Exception('require var if_exists: {if_exists} ')

    client = bigquery.Client()
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    table_id = f"{project_id}.{dataset_id}.{table_id}"

    job_config = bigquery.LoadJobConfig()

    if if_exists == "truncate":
        job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
    elif if_exists == "append":
        job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND

    if schema:
        job_config.schema = schema
    else:
        job_config.autodetect = True

    try:
        job = client.load_table_from_dataframe(
            df, table_id, job_config=job_config
        )
        job.result()
        logging.info(f"load Table {table_id} from dataframe")
    except Conflict:
        logging.info(f"Provided Schema does not match Table {table_id}")



import gcsfs
from google.cloud import storage

from etl_bq_tools.utils.memory import get_reduce_memory


def gs_bucket_list_patterns(project_id, bucket_name,
                            prefix_start=None, prefix_end=None,
                            logging=None, key_path=None):
    """
    :param project_id: String
    :param bucket_name: String
    :param prefix_start: String
    :param prefix_end: String
    :param logging: Object
    :param key_path: file.json
    :return: list_blobs
    """
    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not bucket_name:
        raise Exception('require var bucket_name:{bucket_name} ')

    client = storage.Client(project=project_id)
    if project_id and key_path:
        client = storage.Client.from_service_account_json(key_path)

    try:
        bucket = client.get_bucket(bucket_name)
        blobs = bucket.list_blobs()

        filenames = list()
        if prefix_start:
            filenames = list(sorted([str(blob.name) for blob in blobs
                                     if str(blob.name).startswith(prefix_start)]))

        elif prefix_end:
            filenames = list(sorted([str(blob.name) for blob in blobs
                                     if str(blob.name).endswith(prefix_end)]))

        elif prefix_start and prefix_start:
            filenames = list(sorted([str(blob.name) for blob in blobs
                                     if str(blob.name).startswith(prefix_start) and
                                     str(blob.name).endswith(prefix_end)]))

        elif not prefix_start and not prefix_start:
            filenames = list(sorted([str(blob.name) for blob in blobs]))

        logging.info(f"list_blobs from  bucket {bucket_name} ")
        return filenames
    except Exception as e:
        logging.info(f"BucketName Not Found {bucket_name}")


def gs_bucket_list_to_df(project_id, bucket_name,
                         filename=None, filename_col=None,
                         logging=None, key_path=None):
    import pandas as pd
    import gc

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not bucket_name:
        raise Exception('require var bucket_name:{bucket_name} ')
    if not filename:
        raise Exception('require var filename:{filename} ')
    if not filename_col:
        raise Exception('require var filename_col:{filename_col} ')
    if not key_path:
        raise Exception('require var key_path:{key_path} ')

    try:
        fs = gcsfs.GCSFileSystem(project=project_id,
                                 token=key_path)

        with fs.open(f"gs://{bucket_name}/{filename}") as f:
            frame = pd.read_csv(f)
            frame[f'{filename_col}'] = filename
            df_columns = frame.columns
            frame[df_columns] = frame[df_columns].astype(str)
            df2 = get_reduce_memory(frame, False)
            f.close()
        del frame
        gc.collect()

        logging.info(f"blobs from dataframe {bucket_name}/{filename} ")
        return df2
    except Exception as e:
        logging.info(f"BucketName Not Found {bucket_name} ")



from etl_bq_tools.functions.bigquery import bq_copy_table
from etl_bq_tools.functions.bigquery import bq_create_dataset
from etl_bq_tools.functions.bigquery import bq_create_table
from etl_bq_tools.functions.bigquery import bq_delete_dataset
from etl_bq_tools.functions.bigquery import bq_delete_table
from etl_bq_tools.functions.bigquery import bq_list_dataset
from etl_bq_tools.functions.bigquery import bq_list_tables
from etl_bq_tools.functions.bigquery import bq_query_to_df
from etl_bq_tools.functions.bq import bq_execute_sql
from etl_bq_tools.functions.bq import bq_set_project
from etl_bq_tools.functions.cloud_storage import gs_bucket_list_patterns
from etl_bq_tools.functions.cloud_storage import gs_bucket_list_to_df
from etl_bq_tools.functions.dataframe import upload_df_to_bq
from etl_bq_tools.utils.logger import get_logger
from etl_bq_tools.utils.memory import get_reduce_memory
from etl_bq_tools.utils.time_execution import get_time_function_execution

__all__ = ["bq_copy_table",
           "bq_create_dataset",
           "bq_create_table",
           "bq_delete_dataset",
           "bq_delete_table",
           "bq_list_dataset",
           "bq_list_tables",
           "bq_execute_sql",
           "bq_set_project",
           "bq_query_to_df",
           "upload_df_to_bq",
           "get_reduce_memory",
           "get_logger",
           "gs_bucket_list_patterns",
           "gs_bucket_list_to_df",
           "get_time_function_execution"]
